
!> @file Overlap.f90
!!
!! Defines utility routines for computing overlap
!! integrals between gaussian basis elements.
!!
!! Author: I. Duchemin July 2015
!!


!> Two centers overlap integral for Cubic Spherical Harmonics (C)
!!
!!  \f$
!!      \int dr \, Y_{xyz}^{(1)}(r-R_1) Y_{xyz}^{(2)}(r-R_2)
!!  \f$
!!  
!!  nx+ny+nz must stay <= 6 for each spherical harmonics
!!  
recursive function C_Overlap_C(alpha1,r1,nx1,ny1,nz1,alpha2,r2,nx2,ny2,nz2)
  
  use mod_R_1_norm
  use mod_XZY_power_to_ir
  use mod_CubicHarmonicsProduct
  
  implicit none
  
  ! inputs arguments
  real(kind=8) :: r1(3)    !< center for first cubic Harmonic
  real(kind=8) :: r2(3)    !< center for second cubic Harmonic
  real(kind=8) :: alpha1   !< exponent for first cubic Harmonic
  real(kind=8) :: alpha2   !< exponent for second cubic Harmonic
  integer      :: nx1      !< x power for first cubic Harmonic
  integer      :: ny1      !< y power for first cubic Harmonic
  integer      :: nz1      !< z power for first cubic Harmonic
  integer      :: nx2      !< x power for second cubic Harmonic
  integer      :: ny2      !< y power for second cubic Harmonic
  integer      :: nz2      !< z power for second cubic Harmonic
  
  ! return value
  real(kind=8)  :: C_Overlap_C
  
  ! local variable
  integer       :: l3
  integer       :: ir1
  integer       :: ir2
  real(kind=8)  :: c1(455)
  real(kind=8)  :: c2(455)
  real(kind=8)  :: c3(455)
  real(kind=8)  :: r3(3)
  real(kind=8)  :: a3
  
  ! get 1st member index in the R basis
  ir1 =ir_index(nx1,ny1,nz1)
  
  ! get 2nd member index in the R basis
  ir2 =ir_index(nx2,ny2,nz2)
  
  ! set 1st member coeff in the R basis
  c1=0.0d0
  c1(ir1)=1.0d0
  
  ! set 2nd member coeff in the R basis
  c2=0.0d0
  c2(ir2)=1.0d0
  
  ! get product of the two cubic harmonics
  call RxR_to_R(alpha1,r1,c1,nx1+ny1+nz1,alpha2,r2,c2,nx2+ny2+nz2,a3,r3,c3,l3)
  
  ! compute 1 norm of decomposition
  C_Overlap_C =R_1_norm(a3,c3,l3)
  
end function
  
  
!> Two centers overlap integral for Solid Spherical Harmonics (Y)
!!
!!  \f$
!!      \int dr \, Y_{lm}^{(1)}(r-R_1) Y_{lm}^{(2)}(r-R_2)
!!  \f$
!!  
!!  l must stay <= 6 for each spherical harmonics
!!  
recursive function Y_Overlap_Y(alpha1,r1,l1,m1,alpha2,r2,l2,m2)
  
  use mod_R_1_norm
  use mod_R_from_Y
  use mod_CubicHarmonicsProduct
  
  implicit none
  
  ! inputs arguments
  real(kind=8) :: r1(3)    !< center for first spherical Harmonic
  real(kind=8) :: r2(3)    !< center for second spherical Harmonic
  real(kind=8) :: alpha1   !< exponent for first spherical Harmonic
  real(kind=8) :: alpha2   !< exponent for second spherical Harmonic
  integer      :: l1       !< angular momentum for first spherical Harmonic
  integer      :: l2       !< angular momentum for second spherical Harmonic
  integer      :: m1       !< orbital momentum for first spherical Harmonic
  integer      :: m2       !< orbital momentum for second spherical Harmonic
  
  ! return value
  real(kind=8)  :: Y_Overlap_Y
  
  ! local variable
  integer       :: l3
  real(kind=8)  :: c1(455)
  real(kind=8)  :: c2(455)
  real(kind=8)  :: c3(455)
  real(kind=8)  :: r3(3)
  real(kind=8)  :: a3
  
  ! decomposition of the first solid harmonic into cubic harmonic
  call R_from_Y(l1,m1,c1)
  
  ! decomposition of the second solid harmonic into cubic harmonic
  call R_from_Y(l2,m2,c2)
  
  ! get product of the two cubic harmonics
  call RxR_to_R(alpha1,r1,c1,l1,alpha2,r2,c2,l2,a3,r3,c3,l3)
  
  ! compute 1 norm of decomposition
  Y_Overlap_Y =R_1_norm(a3,c3,l3)
  
end function

!> Three centers overlap integral for Cubic Spherical Harmonics (C)
!!
!!  \f$
!!      \int dr \, Y_{xyz}^{(1)}(r-R_1) Y_{xyz}^{(2)}(r-R_2) Y_{xyz}^{(3)}(r-R_3)
!!  \f$
!!  
!!  l_tot must stay <= 14
!!  
recursive function CC_Overlap_C(alpha1,r1,nx1,ny1,nz1,alpha2,r2,nx2,ny2,nz2,alpha3,r3,nx3,ny3,nz3)
  
  use mod_R_1_norm
  use mod_XZY_power_to_ir
  use mod_CubicHarmonicsProduct
  
  implicit none
  
  ! inputs arguments
  real(kind=8) :: r1(3)    !< center for first cubic Harmonic
  real(kind=8) :: r2(3)    !< center for second cubic Harmonic
  real(kind=8) :: r3(3)    !< center for third cubic Harmonic
  real(kind=8) :: alpha1   !< exponent for first cubic Harmonic
  real(kind=8) :: alpha2   !< exponent for second cubic Harmonic
  real(kind=8) :: alpha3   !< exponent for third cubic Harmonic
  integer      :: nx1      !< x power for first cubic Harmonic
  integer      :: ny1      !< y power for first cubic Harmonic
  integer      :: nz1      !< z power for first cubic Harmonic
  integer      :: nx2      !< x power for second cubic Harmonic
  integer      :: ny2      !< y power for second cubic Harmonic
  integer      :: nz2      !< z power for second cubic Harmonic
  integer      :: nx3      !< x power for third cubic Harmonic
  integer      :: ny3      !< y power for third cubic Harmonic
  integer      :: nz3      !< z power for third cubic Harmonic
  
  ! return value
  real(kind=8)  :: CC_Overlap_C
  
  ! local variable
  integer       :: l4
  integer       :: l5
  integer       :: ir1
  integer       :: ir2
  real(kind=8)  :: c1(680)
  real(kind=8)  :: c2(680)
  real(kind=8)  :: c3(680)
  real(kind=8)  :: r4(3)
  real(kind=8)  :: r5(3)
  real(kind=8)  :: a4
  real(kind=8)  :: a5
  
  ! get 1st member index in the R basis
  ir1 =ir_index(nx1,ny1,nz1)
  
  ! get 2nd member index in the R basis
  ir2 =ir_index(nx2,ny2,nz2)
  
  ! set 1st member coeff in the R basis
  c1=0.0d0
  c1(ir1)=1.0d0
  
  ! set 2nd member coeff in the R basis
  c2=0.0d0
  c2(ir2)=1.0d0
  
  ! get product of the two cubic harmonics
  call RxR_to_R(alpha1,r1,c1,nx1+ny1+nz1,alpha2,r2,c2,nx2+ny2+nz2,a4,r4,c3,l4)
  
  ! get 3rd member index in the R basis
  ir1 =ir_index(nx3,ny3,nz3)
  
  ! set 3rd member coeff in the R basis
  c1=0.0d0
  c1(ir1)=1.0d0
  
  ! get product of the three cubic harmonics
  call RxR_to_R(a4,r4,c3,l4,alpha3,r3,c1,nx3+ny3+nz3,a5,r5,c2,l5)
  
  ! compute 1 norm of decomposition
  CC_Overlap_C =R_1_norm(a5,c2,l5)
  
end function

!> Three centers overlap integral for Solid Spherical Harmonics (Y)
!!
!!  \f$
!!      \int dr \, Y_{lm}^{(1)}(r-R_1) Y_{lm}^{(2)}(r-R_2) Y_{lm}^{(3)}(r-R_3)
!!  \f$
!!  
!!  l_tot must stay <= 14
!!  
recursive function YY_Overlap_Y(alpha1,r1,l1,m1,alpha2,r2,l2,m2,alpha3,r3,l3,m3)
  
  use mod_R_1_norm
  use mod_R_from_Y
  use mod_CubicHarmonicsProduct
  
  implicit none
  
  ! inputs arguments
  real(kind=8) :: r1(3)    !< center for first spherical Harmonic
  real(kind=8) :: r2(3)    !< center for second spherical Harmonic
  real(kind=8) :: r3(3)    !< center for third spherical Harmonic
  real(kind=8) :: alpha1   !< exponent for first spherical Harmonic
  real(kind=8) :: alpha2   !< exponent for second spherical Harmonic
  real(kind=8) :: alpha3   !< exponent for third spherical Harmonic
  integer      :: l1       !< angular momentum for first spherical Harmonic
  integer      :: l2       !< angular momentum for second spherical Harmonic
  integer      :: l3       !< angular momentum for third spherical Harmonic
  integer      :: m1       !< orbital momentum for first spherical Harmonic
  integer      :: m2       !< orbital momentum for second spherical Harmonic
  integer      :: m3       !< orbital momentum for third spherical Harmonic
  
  ! return value
  real(kind=8)  :: YY_Overlap_Y
  
  ! local variable
  integer       :: l4
  integer       :: l5
  real(kind=8)  :: c1(680)
  real(kind=8)  :: c2(680)
  real(kind=8)  :: c3(680)
  real(kind=8)  :: r4(3)
  real(kind=8)  :: r5(3)
  real(kind=8)  :: a4
  real(kind=8)  :: a5
  
  ! decomposition of the first solid harmonic into cubic harmonic
  call R_from_Y(l1,m1,c1)
  
  ! decomposition of the second solid harmonic into cubic harmonic
  call R_from_Y(l2,m2,c2)
  
  ! get product of the two cubic harmonics
  call RxR_to_R(alpha1,r1,c1,l1,alpha2,r2,c2,l2,a4,r4,c3,l4)
  
  ! decomposition of the first solid harmonic into cubic harmonic
  call R_from_Y(l3,m3,c1)
  
  ! get product of the two cubic harmonics
  call RxR_to_R(alpha3,r3,c1,l3,a4,r4,c3,l4,a5,r5,c2,l5)
  
  ! compute 1 norm of decomposition
  YY_Overlap_Y =R_1_norm(a5,c2,l5)
  
end function


