/*   EXTRAITS DE LA LICENCE
	Copyright CEA, contributeurs : Damien
	CALISTE, laboratoire L_Sim, (2017)
  
	Adresses mèl :
	CALISTE, damien P caliste AT cea P fr.

	Ce logiciel est un programme informatique servant à visualiser des
	structures atomiques dans un rendu pseudo-3D. 

	Ce logiciel est régi par la licence CeCILL soumise au droit français et
	respectant les principes de diffusion des logiciels libres. Vous pouvez
	utiliser, modifier et/ou redistribuer ce programme sous les conditions
	de la licence CeCILL telle que diffusée par le CEA, le CNRS et l'INRIA 
	sur le site "http://www.cecill.info".

	Le fait que vous puissiez accéder à cet en-tête signifie que vous avez 
	pris connaissance de la licence CeCILL, et que vous en avez accepté les
	termes (cf. le fichier Documentation/licence.fr.txt fourni avec ce logiciel).
*/

/*   LICENCE SUM UP
	Copyright CEA, contributors : Damien
	CALISTE, laboratoire L_Sim, (2017)

	E-mail addresses :
	CALISTE, damien P caliste AT cea P fr.

	This software is a computer program whose purpose is to visualize atomic
	configurations in 3D.

	This software is governed by the CeCILL  license under French law and
	abiding by the rules of distribution of free software.  You can  use, 
	modify and/ or redistribute the software under the terms of the CeCILL
	license as circulated by CEA, CNRS and INRIA at the following URL
	"http://www.cecill.info". 

	The fact that you are presently reading this means that you have had
	knowledge of the CeCILL license and that you accept its terms. You can
	find a copy of this licence shipped with this software at Documentation/licence.en.txt.
*/

#include <coreTools/toolFiles.h>

#ifdef HAVE_LIB_ARCHIVE
#include <archive.h>
#include <string.h>
#endif

/**
 * SECTION:toolFiles
 * @short_description: abstract reading of compressed or not text files.
 *
 * <para>This class defines objects to read text files with the same
 * API than with #GIOChannel. The addition is that these files can be
 * transparently compressed or not.</para>
 */

struct _ToolFilesPrivate
{
  gboolean dispose_has_run;

  gchar *filename;
  
#ifdef HAVE_LIB_ARCHIVE
  struct archive *archive;
#define BUFF_SIZE 1024
  char buff[BUFF_SIZE + 1];
  char *cur;
#endif

  GIOChannel *direct;
};

static void tool_files_dispose(GObject* obj);
static void tool_files_finalize(GObject* obj);

G_DEFINE_TYPE(ToolFiles, tool_files, VISU_TYPE_OBJECT)

static void tool_files_class_init(ToolFilesClass *klass)
{
  DBG_fprintf(stderr, "ToolFiles: creating the class of the object.\n");

  /* DBG_fprintf(stderr, "                - adding new signals ;\n"); */

  /* Connect freeing methods. */
  G_OBJECT_CLASS(klass)->dispose  = tool_files_dispose;
  G_OBJECT_CLASS(klass)->finalize = tool_files_finalize;

  g_type_class_add_private(klass, sizeof(ToolFilesPrivate));
}

static void tool_files_init(ToolFiles *obj)
{
  DBG_fprintf(stderr, "ToolFiles: creating a new file reader (%p).\n", (gpointer)obj);

  obj->priv = G_TYPE_INSTANCE_GET_PRIVATE(obj, TOOL_TYPE_FILES,
                                          ToolFilesPrivate);

  obj->priv->dispose_has_run = FALSE;
  obj->priv->filename = (gchar*)0;
#ifdef HAVE_LIB_ARCHIVE
  obj->priv->archive = (struct archive*)0;
  obj->priv->buff[BUFF_SIZE] = '\0';
  obj->priv->cur = (char*)0;
#endif
  obj->priv->direct = (GIOChannel*)0;
}
static void tool_files_dispose(GObject* obj)
{
  ToolFiles *self = TOOL_FILES(obj);
  DBG_fprintf(stderr, "ToolFiles: dispose object %p.\n", (gpointer)obj);

  if (self->priv->dispose_has_run)
    return;
  self->priv->dispose_has_run = TRUE;

  if (self->priv->direct)
    g_io_channel_unref(self->priv->direct);

  /* Chain up to the parent class */
  G_OBJECT_CLASS(tool_files_parent_class)->dispose(obj);
}
static void tool_files_finalize(GObject* obj)
{
  ToolFiles *self = TOOL_FILES(obj);
  DBG_fprintf(stderr, "ToolFiles: finalize object %p.\n", (gpointer)obj);

  g_free(self->priv->filename);
#ifdef HAVE_LIB_ARCHIVE
  if (self->priv->archive)
    archive_read_free(self->priv->archive);
#endif

  /* Chain up to the parent class */
  G_OBJECT_CLASS(tool_files_parent_class)->finalize(obj);
}

/**
 * tool_files_new:
 *
 * Creates a new #ToolFiles object.
 *
 * Since: 3.8
 *
 * Returns: (transfer full): a newly created #ToolFiles object.
 */
ToolFiles* tool_files_new()
{
  return TOOL_FILES(g_object_new(TOOL_TYPE_FILES, NULL));
}

/**
 * tool_files_open:
 * @file: a #ToolFiles object.
 * @filename: a filename.
 * @error: a location for an error.
 *
 * Open @filename for read access. The file can be compressed or not.
 *
 * Since: 3.8
 *
 * Returns: TRUE if no error occured when opening the file.
 **/
gboolean tool_files_open(ToolFiles *file, const gchar *filename, GError **error)
{
#ifdef HAVE_LIB_ARCHIVE
  int r;
  struct archive_entry *ae;
#endif

  g_return_val_if_fail(TOOL_IS_FILES_TYPE(file), FALSE);
  g_return_val_if_fail(!error || *error == (GError*)0, FALSE);

  file->priv->filename = g_strdup(filename);
#ifdef HAVE_LIB_ARCHIVE
  file->priv->archive = archive_read_new();
  archive_read_support_filter_all(file->priv->archive);
  archive_read_support_format_raw(file->priv->archive);
  
  r = archive_read_open_filename(file->priv->archive, filename, 16384);
  if (r == ARCHIVE_OK)
    {
      r = archive_read_next_header(file->priv->archive, &ae);
      if (r != ARCHIVE_OK)
        {
          g_set_error(error, G_FILE_ERROR, G_FILE_ERROR_FAILED,
                      _("cannot read next archive entry.\n"));
          return FALSE;
        }
      return TRUE;
    }
  else
    {
      archive_read_free(file->priv->archive);
      file->priv->archive = (struct archive*)0;
    }
#endif

  file->priv->direct = g_io_channel_new_file(filename, "r", error);
  if (!file->priv->direct)
    return FALSE;

  return TRUE;
}

/**
 * tool_files_read_line_string:
 * @file: a #ToolFiles object.
 * @buffer: (out caller-allocates): an allocated string.
 * @terminator_pos: (out caller-allocates): position of the terminator
 * in @buffer.
 * @error: an error location.
 *
 * Read a new line from @file and put it into @buffer. This works like
 * g_io_channel_read_line_string() but is transparent for compressed files.
 *
 * Since: 3.8
 *
 * Returns: a status.
 **/
GIOStatus tool_files_read_line_string(ToolFiles *file, GString *buffer,
                                      gsize *terminator_pos, GError **error)
{
#ifdef HAVE_LIB_ARCHIVE
  ssize_t size;
  char *eol;
#endif
  
  g_return_val_if_fail(TOOL_IS_FILES_TYPE(file), G_IO_STATUS_ERROR);
  g_return_val_if_fail(!error || *error == (GError*)0, G_IO_STATUS_ERROR);

#ifdef HAVE_LIB_ARCHIVE
  if (file->priv->archive)
    {
      g_string_set_size(buffer, 0);

      while (TRUE)
        {
          if (file->priv->cur)
            {
              eol = strchr(file->priv->cur, '\n');
              if (eol)
                {
                  *eol = '\0';
                  g_string_append(buffer, file->priv->cur);
                  file->priv->cur = eol + 1;
          
                  return G_IO_STATUS_NORMAL;
                }
              g_string_append(buffer, file->priv->cur);
            }
      
          file->priv->cur = (char*)0;
          size = archive_read_data(file->priv->archive, file->priv->buff, BUFF_SIZE);
          if (size < 0)
            {
              g_set_error(error, G_FILE_ERROR, G_FILE_ERROR_IO,
                          _("read error from archive.\n"));
              return G_IO_STATUS_ERROR;
            }
          if (size == 0)
            return buffer->len ? G_IO_STATUS_NORMAL : G_IO_STATUS_EOF;
          file->priv->buff[size] = '\0';
          file->priv->cur = file->priv->buff;
        }

      return G_IO_STATUS_NORMAL;
    }
#endif
  if (file->priv->direct)
    {
      return g_io_channel_read_line_string(file->priv->direct, buffer,
                                           terminator_pos, error);
    }
  g_set_error(error, G_FILE_ERROR, G_FILE_ERROR_NOENT,
              _("file not opened.\n"));
  return G_IO_STATUS_ERROR;
}

/**
 * tool_files_rewind:
 * @file: a #ToolFiles object.
 * @error: an error location.
 *
 * Transparently rewind @file at the beginning for compressed files or not.
 *
 * Since: 3.8
 *
 * Returns: a status.
 **/
GIOStatus tool_files_rewind(ToolFiles *file, GError **error)
{
#ifdef HAVE_LIB_ARCHIVE
  if (file->priv->archive)
    {
      /* Close and reopen. */
      archive_read_free(file->priv->archive);
      return tool_files_open(file, file->priv->filename, error);
    }
#endif
  if (file->priv->direct)
    {
      return g_io_channel_seek_position(file->priv->direct, 0, G_SEEK_SET, error);
    }
  g_set_error(error, G_FILE_ERROR, G_FILE_ERROR_NOENT,
              _("file not opened.\n"));
  return G_IO_STATUS_ERROR;
}
