/*   EXTRAITS DE LA LICENCE
	Copyright CEA, contributeurs : Damien
	CALISTE, laboratoire L_Sim, (2016)
  
	Adresse mèl :
	CALISTE, damien P caliste AT cea P fr.

	Ce logiciel est un programme informatique servant à visualiser des
	structures atomiques dans un rendu pseudo-3D. 

	Ce logiciel est régi par la licence CeCILL soumise au droit français et
	respectant les principes de diffusion des logiciels libres. Vous pouvez
	utiliser, modifier et/ou redistribuer ce programme sous les conditions
	de la licence CeCILL telle que diffusée par le CEA, le CNRS et l'INRIA 
	sur le site "http://www.cecill.info".

	Le fait que vous puissiez accéder à cet en-tête signifie que vous avez 
	pris connaissance de la licence CeCILL, et que vous en avez accepté les
	termes (cf. le fichier Documentation/licence.fr.txt fourni avec ce logiciel).
*/

/*   LICENCE SUM UP
	Copyright CEA, contributors : Damien
	CALISTE, laboratoire L_Sim, (2016)

	E-mail address:
	CALISTE, damien P caliste AT cea P fr.

	This software is a computer program whose purpose is to visualize atomic
	configurations in 3D.

	This software is governed by the CeCILL  license under French law and
	abiding by the rules of distribution of free software.  You can  use, 
	modify and/ or redistribute the software under the terms of the CeCILL
	license as circulated by CEA, CNRS and INRIA at the following URL
	"http://www.cecill.info". 

	The fact that you are presently reading this means that you have had
	knowledge of the CeCILL license and that you accept its terms. You can
	find a copy of this licence shipped with this software at Documentation/licence.en.txt.
*/

#include "visu_dataatomic.h"

#include "visu_basic.h"

#include "loaders/atomic_ascii.h"
#include "loaders/atomic_d3.h"
#include "loaders/atomic_xyz.h"
#include "loaders/atomic_yaml.h"

/**
 * SECTION: visu_dataatomic
 * @short_description: a class of nodes representing atomic data and
 * providing associated loading methods.
 *
 * <para>This class provides #VisuDataLoader for atomic data representation.</para>
 */

enum
  {
    PROP_0,
    PROP_LABEL,
    N_PROP
  };
/* static GParamSpec *_properties[N_PROP]; */

struct _VisuDataAtomicPrivate
{
  gchar *file;
  VisuDataLoader *format;
};

static GList *_atomicFormats = NULL;
static void _initAtomicFormats(void)
{
  visu_data_atomic_class_addLoader(visu_data_loader_ascii_getStatic());
  visu_data_atomic_class_addLoader(visu_data_loader_d3_getStatic());
  visu_data_atomic_class_addLoader(visu_data_loader_xyz_getStatic());
  if (visu_data_loader_yaml_getStatic())
    visu_data_atomic_class_addLoader(visu_data_loader_yaml_getStatic());
}

static void visu_data_atomic_finalize    (GObject* obj);
static void visu_data_atomic_get_property(GObject* obj, guint property_id,
                                          GValue *value, GParamSpec *pspec);
static gboolean visu_data_atomic_load(VisuDataLoadable *self, guint iSet,
                                      GCancellable *cancel, GError **error);
static void visu_data_atomic_addFilename(VisuDataLoadable *self, const gchar *filename,
                                         guint fileType, VisuDataLoader *format);
static const gchar* visu_data_atomic_getFilename(const VisuDataLoadable *self,
                                                 guint fileType, VisuDataLoader **format);

G_DEFINE_TYPE(VisuDataAtomic, visu_data_atomic, VISU_TYPE_DATA_LOADABLE)

static void visu_data_atomic_class_init(VisuDataAtomicClass *klass)
{
  /* Connect the overloading methods. */
  G_OBJECT_CLASS(klass)->finalize     = visu_data_atomic_finalize;
  G_OBJECT_CLASS(klass)->get_property = visu_data_atomic_get_property;
  VISU_DATA_LOADABLE_CLASS(klass)->load = visu_data_atomic_load;
  VISU_DATA_LOADABLE_CLASS(klass)->addFilename = visu_data_atomic_addFilename;
  VISU_DATA_LOADABLE_CLASS(klass)->getFilename = visu_data_atomic_getFilename;

  g_object_class_override_property(G_OBJECT_CLASS(klass), PROP_LABEL, "label");

  g_type_class_add_private(klass, sizeof(VisuDataAtomicPrivate));

  _initAtomicFormats();
}
static void visu_data_atomic_init(VisuDataAtomic *obj)
{
  obj->priv = G_TYPE_INSTANCE_GET_PRIVATE(obj, VISU_TYPE_DATA_ATOMIC,
                                          VisuDataAtomicPrivate);

  /* Private data. */
  obj->priv->file    = (gchar*)0;
  obj->priv->format  = (VisuDataLoader*)0;
}
static void visu_data_atomic_finalize(GObject* obj)
{
  VisuDataAtomic *data;

  data = VISU_DATA_ATOMIC(obj);

  g_free(data->priv->file);

  /* Chain up to the parent class */
  G_OBJECT_CLASS(visu_data_atomic_parent_class)->finalize(obj);
}
static void visu_data_atomic_get_property(GObject* obj, guint property_id,
                                          GValue *value, GParamSpec *pspec)
{
  VisuDataAtomic *self = VISU_DATA_ATOMIC(obj);

  switch (property_id)
    {
    case PROP_LABEL:
      g_value_take_string(value, g_path_get_basename(self->priv->file));
      break;
    default:
      /* We don't have any other property... */
      G_OBJECT_WARN_INVALID_PROPERTY_ID(obj, property_id, pspec);
      break;
    }
}

/**
 * visu_data_atomic_new:
 *
 * Creates a #VisuDataAtomic object.
 *
 * Since: 3.8
 *
 * Returns: a newly allocated #VisuDataAtomic object.
 **/
VisuDataAtomic* visu_data_atomic_new(void)
{
  return g_object_new(VISU_TYPE_DATA_ATOMIC, "n-files", 1, NULL);
}

/**
 * visu_data_atomic_new_withFile:
 * @file: a filename.
 *
 * Creates a #VisuDataAtomic object and set @file to it.
 *
 * Since: 3.8
 *
 * Returns: a newly allocated #VisuDataAtomic object.
 **/
VisuDataAtomic* visu_data_atomic_new_withFile(const gchar *file)
{
  VisuDataAtomic *data;

  data = g_object_new(VISU_TYPE_DATA_ATOMIC, "n-files", 1, NULL);
  data->priv->file = tool_path_normalize(file);

  return data;
}

/**
 * visu_data_atomic_class_addLoader:
 * @loader: (transfer full): a #VisuDataLoader object.
 *
 * Add @loader to the list of #VisuDataLoader to be used when
 * visu_data_loadable_load() is called.
 *
 * Since: 3.8
 **/
void visu_data_atomic_class_addLoader(VisuDataLoader *loader)
{
  if (g_list_find(_atomicFormats, loader))
    return;

  tool_file_format_setPropertiesFromCLI(TOOL_FILE_FORMAT(loader));
  _atomicFormats = g_list_prepend(_atomicFormats, loader);
  _atomicFormats = g_list_sort(_atomicFormats,
                               (GCompareFunc)visu_data_loader_comparePriority);
}

/**
 * visu_data_atomic_class_getLoaders:
 *
 * Returns a list of available #VisuDataLoader.
 *
 * Since: 3.8
 *
 * Returns: (transfer none) (element-type VisuDataLoader): a list of
 * #VisuDataLoader owned by V_Sim.
 **/
GList* visu_data_atomic_class_getLoaders(void)
{
  return _atomicFormats;
}
/**
 * visu_data_atomic_class_getFileDescription:
 *
 * Returns a translated string describing what is files loaded by
 * #VisuDataAtomic objects.
 *
 * Since: 3.8
 *
 * Returns: a string owned by V_Sim.
 **/
const gchar* visu_data_atomic_class_getFileDescription(void)
{
  return _("Position files");
}
/**
 * visu_data_atomic_class_finalize:
 *
 * Empty the list of known loaders.
 *
 * Since: 3.8
 **/
void visu_data_atomic_class_finalize(void)
{
  g_list_free_full(_atomicFormats, (GDestroyNotify)g_object_unref);
  _atomicFormats = (GList*)0;
}

/**
 * visu_data_atomic_setFile:
 * @data: a #VisuDataAtomic object.
 * @filename: a filename.
 * @format: (allow-none): the associated #VisuDataLoader format, if
 * any.
 *
 * Associate @filename to @data. When visu_data_loadable_load() is
 * called and @format provided, this #VisuDataLoader will be used
 * only, otherwise the list of available loaders (see
 * visu_data_atomic_class_addLoader()) will be used.
 *
 * Since: 3.8
 **/
void visu_data_atomic_setFile(VisuDataAtomic *data, const gchar *filename,
                              VisuDataLoader *format)
{
  g_return_if_fail(VISU_IS_DATA_ATOMIC(data));

  data->priv->format = format;
  g_free(data->priv->file);
  data->priv->file = tool_path_normalize(filename);
}
static void visu_data_atomic_addFilename(VisuDataLoadable *self, const gchar *filename,
                                         guint fileType, VisuDataLoader *format)
{
  g_return_if_fail(fileType == 0);

  visu_data_atomic_setFile(VISU_DATA_ATOMIC(self), filename, format);
}
/**
 * visu_data_atomic_getFile:
 * @data: a #VisuDataAtomic object.
 * @format: (out caller-allocates): a location to store a
 * #VisuDataLoader object.
 *
 * Returns the file defined in @data and its associated @format, if any.
 *
 * Since: 3.8
 *
 * Returns: a string owned by V_Sim.
 **/
const gchar* visu_data_atomic_getFile(VisuDataAtomic *data, VisuDataLoader **format)
{
  g_return_val_if_fail(VISU_IS_DATA_ATOMIC(data), (const gchar*)0);

  if (format)
    *format = data->priv->format;
  return data->priv->file;
}
static const gchar* visu_data_atomic_getFilename(const VisuDataLoadable *self,
                                                 guint fileType, VisuDataLoader **format)
{
  g_return_val_if_fail(fileType == 0, (const gchar*)0);

  return visu_data_atomic_getFile(VISU_DATA_ATOMIC(self), format);
}

static gboolean _finalize(VisuDataAtomic *data, guint iSet)
{
  VisuNodeArrayIter iter;
  ToolUnits unit, preferedUnit;
  VisuBox *box;

  /* We do a last check on population... in case. */
  visu_node_array_iter_new(VISU_NODE_ARRAY(data), &iter);
  g_return_val_if_fail(iter.nAllStoredNodes, FALSE);

  DBG_fprintf(stderr, "Data Atomic: loading OK (%d nodes).\n", iter.nAllStoredNodes);
  visu_node_array_populationIncrease(VISU_NODE_ARRAY(data), (GArray*)0);

  /* Setup additional infos. */
  visu_data_setISubset(VISU_DATA(data), iSet);

  box = visu_boxed_getBox(VISU_BOXED(data));
  unit = visu_box_getUnit(box);
  preferedUnit = visu_basic_getPreferedUnit();
  if (preferedUnit != TOOL_UNITS_UNDEFINED &&
      unit != TOOL_UNITS_UNDEFINED &&
      unit != preferedUnit)
    visu_box_setUnit(box, preferedUnit);

  return TRUE;
}

static gboolean visu_data_atomic_load(VisuDataLoadable *self, guint iSet,
                                      GCancellable *cancel, GError **error)
{
  VisuDataAtomic *data;
  GList *lst;
  VisuDataLoader *loader;

  g_return_val_if_fail(VISU_IS_DATA_ATOMIC(self), FALSE);

  data = VISU_DATA_ATOMIC(self);

  if (!visu_data_loadable_checkFile(self, 0, error))
    return FALSE;

  for (lst = _atomicFormats; lst; lst = g_list_next(lst))
    {
      loader = VISU_DATA_LOADER(lst->data);

      /* Each load may set error even if the format is not recognise
	 and loadOK is FALSE, then we need to free the error. */
      g_clear_error(error);

      if (!data->priv->format || loader == data->priv->format)
	{
	  DBG_fprintf(stderr,"Data Atomic: testing '%s' with format: %s.\n",
		      data->priv->file, tool_file_format_getName(TOOL_FILE_FORMAT(loader)));
	  if (visu_data_loader_load(loader, data->priv->file, VISU_DATA(self),
                                    iSet, cancel, error))
            return (*error) ? FALSE : _finalize(data, iSet);
          if (*error && (*error)->domain == G_FILE_ERROR)
            return FALSE;
        }
    }
  g_clear_error(error);
  g_set_error(error, VISU_ERROR_DATA_LOADABLE, DATA_LOADABLE_ERROR_UNKNOWN,
              _("Impossible to load '%s', unrecognised format.\n"), data->priv->file);
  return FALSE;
}

/**
 * visu_data_atomic_getForces:
 * @dataObj: a #VisuDataAtomic object.
 * @create: a boolean.
 *
 * Retrieves the #VisuNodeValuesVector used to store forces for
 * @dataObj, create it depending on @create if not exists.
 *
 * Since: 3.8
 *
 * Returns: (transfer none): a #VisuNodeValuesVector object, owned by V_Sim.
 **/
VisuNodeValuesVector* visu_data_atomic_getForces(VisuDataAtomic *dataObj,
                                                 gboolean create)
{
  VisuNodeValuesVector *vect;

  if (!dataObj)
    return (VisuNodeValuesVector*)0;

  vect = (VisuNodeValuesVector*)visu_data_getNodeProperties(VISU_DATA(dataObj),
                                                            _("Forces"));
  if (!vect && create)
    {
      vect = visu_node_values_vector_new(VISU_NODE_ARRAY(dataObj),
                                         _("Forces"));
      visu_node_values_setEditable(VISU_NODE_VALUES(vect), FALSE);
      visu_data_addNodeProperties(VISU_DATA(dataObj), VISU_NODE_VALUES(vect));
    }
  return vect;
}
